// this file is loosely typed as it was copied from another project
import fetch from 'node-fetch';
import queryString from 'query-string';
const { LOG_LEVEL = 'error' } = process.env;
const shouldLog = LOG_LEVEL.includes('alpaca-client');
// "trading" api
const ALPACA_API_PATH_ACCOUNT = '/v2/account';
const ALPACA_API_PATH_ASSETS = '/v2/assets';
const ALPACA_API_PATH_BARS = '/v2/stocks/{symbol}/bars';
const ALPACA_API_PATH_BARS_CRYPTO = '/v1beta2/crypto/bars';
const ALPACA_API_PATH_BARS_MULTI = '/v2/stocks/bars';
const ALPACA_API_PATH_ANNOUNCEMENTS = '/v2/corporate_actions/announcements';
const ALPACA_API_PATH_CALENDAR = '/v2/calendar';
const ALPACA_API_PATH_CLOCK = '/v2/clock';
const ALPACA_API_PATH_ORDERS = '/v2/orders';
const ALPACA_API_PATH_POSITIONS = '/v2/positions';
const ALPACA_API_PATH_PORTFOLIO_HISTORY = '/v2/account/portfolio/history';
const ALPACA_API_PATH_SNAPSHOT = '/v2/stocks/{symbol}/snapshot';
const ALPACA_API_PATH_SNAPSHOT_MULTI = '/v2/stocks/snapshots';
const ALPACA_API_PATH_TRADES = '/v2/stocks/{symbol}/trades';
const ALPACA_API_PATH_TRADES_CRYPTO = '/v1beta2/crypto/trades';
const ALPACA_API_PATH_TRADES_LATEST_CRYPTO = '/v1beta2/crypto/latest/trades';
const ALPACA_API_PATH_WATCHLISTS = '/v2/watchlists';
const getUrlWithQuery = (url, query = {}) => {
    const fullQueryString = !Object.keys(query).length
        ? ''
        : `?${queryString.stringify(query, { encode: false })}`;
    return `${url}${fullQueryString}`;
};
const getErrorJson = (error, metadata, status) => {
    return { error, metadata, status };
};
const getResult = async (response, metadata) => {
    const json = metadata && metadata.type === 'emptyResponse'
        ? { status: response.status }
        : await response.json();
    if (response.status > 399) {
        if (json.message) {
            return getErrorJson(json.message, metadata, response.status);
        }
        return getErrorJson(`an unknown error occurred. status: ${response.status}`, metadata, response.status);
    }
    return json;
};
export default class AlpacaClient {
    constructor(baseUrl, baseUrlData, apiKeyId, secretKey, shouldLogUrl) {
        this.baseUrl = baseUrl;
        this.baseUrlData = baseUrlData;
        this.apiKeyId = apiKeyId;
        this.secretKey = secretKey;
        this.shouldLogUrl = shouldLogUrl;
        this._cache = new Map();
        this.headers = {
            accept: 'application/json',
            'APCA-API-KEY-ID': this.apiKeyId,
            'APCA-API-SECRET-KEY': this.secretKey,
        };
    }
    getCache(cacheKey) {
        const cachedItem = this._cache.get(cacheKey);
        if (shouldLog && cachedItem) {
            console.log(`✔️ response found in cache`, cacheKey);
        }
        return cachedItem;
    }
    // https://alpaca.markets/docs/api-references/trading-api/account/
    async getAccount() {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_ACCOUNT}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/corporate-actions-announcements/
    async getAnnouncements(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrl}${ALPACA_API_PATH_ANNOUNCEMENTS}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, query);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/calendar/
    async getCalendar(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrl}${ALPACA_API_PATH_CALENDAR}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const cachedResponse = this.getCache(apiUrl);
            if (cachedResponse) {
                return cachedResponse;
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            const jsonResponse = await getResult(response);
            if (jsonResponse && !jsonResponse.error) {
                this._cache.set(apiUrl, jsonResponse);
            }
            return jsonResponse;
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, query);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/clock/
    async getClock() {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_CLOCK}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/orders/
    async postOrder(params) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_ORDERS}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'POST',
                headers: this.headers,
                body: JSON.stringify(params),
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { params });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/orders/#get-an-order
    async getOrder({ id }) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_ORDERS}/${id}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { id });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/orders/
    async getOrders(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrl}${ALPACA_API_PATH_ORDERS}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, query);
        }
    }
    async replaceOrder({ id, params }) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_ORDERS}/${id}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'PATCH',
                headers: this.headers,
                body: JSON.stringify(params),
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, params);
        }
    }
    async cancelOrder({ id }) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_ORDERS}/${id}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'DELETE',
                headers: this.headers,
            });
            return getResult(response, {
                type: 'emptyResponse',
            });
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { id });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/portfolio-history/
    async getPortfolioHistory(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrl}${ALPACA_API_PATH_PORTFOLIO_HISTORY}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, query);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/positions/
    async getPosition(symbol) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_POSITIONS}/${symbol}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/positions/
    async getPositions() {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_POSITIONS}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/positions/
    async closePosition(symbol, query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrl}${ALPACA_API_PATH_POSITIONS}/${symbol}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'DELETE',
                headers: this.headers,
            });
            return getResult(response, {
                type: 'emptyResponse',
            });
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/positions/
    async closeAllPositions() {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_POSITIONS}?cancel_orders=true`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'DELETE',
                headers: this.headers,
            });
            return getResult(response, {
                type: 'emptyResponse',
            });
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`);
        }
    }
    // https://alpaca.markets/docs/api-references/market-data-api/stock-pricing-data/historical/#multi-trades
    // and
    // https://alpaca.markets/docs/api-references/market-data-api/crypto-pricing-data/historical/#trades
    async getTrades(symbol, query) {
        try {
            // the way we determine crypto is presence of a `/` in symbol
            const isCrypto = symbol.includes('-') || symbol.includes('/');
            const cryptoSymbol = symbol.replace('-', '/').toUpperCase();
            const apiUrl = !isCrypto
                ? getUrlWithQuery(`${`${this.baseUrlData}${ALPACA_API_PATH_TRADES}`.replace('{symbol}', symbol.toLowerCase())}`, query)
                : getUrlWithQuery(`${this.baseUrlData}${ALPACA_API_PATH_TRADES_CRYPTO}`, {
                    ...query,
                    symbols: cryptoSymbol,
                });
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            const initialResponse = await getResult(response, { symbol, query });
            if (!isCrypto) {
                return initialResponse;
            }
            if (initialResponse.trades) {
                return {
                    symbol,
                    trades: initialResponse.trades[cryptoSymbol],
                    next_page_token: initialResponse.next_page_token,
                };
            }
            else {
                return initialResponse;
            }
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol, query });
        }
    }
    // https://alpaca.markets/docs/api-references/market-data-api/stock-pricing-data/historical/#latest-trade
    async getLatestTrade(symbol) {
        try {
            // the way we determine crypto is presence of a `/` in symbol
            const isCrypto = symbol.includes('-') || symbol.includes('/');
            const cryptoSymbol = symbol.replace('-', '/').toUpperCase();
            const apiUrl = !isCrypto
                ? `${`${this.baseUrlData}${ALPACA_API_PATH_TRADES}/latest`.replace('{symbol}', symbol.toLowerCase())}`
                : getUrlWithQuery(`${this.baseUrlData}${ALPACA_API_PATH_TRADES_LATEST_CRYPTO}`, {
                    symbols: cryptoSymbol,
                });
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            const initialResponse = await getResult(response);
            if (!isCrypto) {
                return initialResponse;
            }
            if (initialResponse.trades) {
                return {
                    symbol: cryptoSymbol,
                    trade: initialResponse.trades[cryptoSymbol],
                };
            }
            else {
                return initialResponse;
            }
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol });
        }
    }
    // https://alpaca.markets/docs/api-references/market-data-api/stock-pricing-data/historical/#bars
    // and
    // https://alpaca.markets/docs/api-references/market-data-api/crypto-pricing-data/historical/#bars
    async getBars(symbol, query) {
        try {
            // the way we determine crypto is presence of a `/` in symbol
            const isCrypto = symbol.includes('-') || symbol.includes('/');
            const cryptoSymbol = symbol.replace('-', '/').toUpperCase();
            const apiUrl = !isCrypto
                ? getUrlWithQuery(`${`${this.baseUrlData}${ALPACA_API_PATH_BARS}`.replace('{symbol}', symbol.toLowerCase())}`, query)
                : getUrlWithQuery(`${this.baseUrlData}${ALPACA_API_PATH_BARS_CRYPTO}`, {
                    ...query,
                    symbols: cryptoSymbol,
                });
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            const initialResponse = await getResult(response, { symbol, query });
            if (!isCrypto) {
                return initialResponse;
            }
            if (initialResponse.bars) {
                return {
                    bars: initialResponse.bars[cryptoSymbol],
                    next_page_token: initialResponse.next_page_token,
                };
            }
            else {
                return initialResponse;
            }
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol, query });
        }
    }
    // https://alpaca.markets/docs/api-references/market-data-api/crypto-pricing-data/historical/#bars
    async getBarsCrypto(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrlData}${ALPACA_API_PATH_BARS_CRYPTO}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response, { query });
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { query });
        }
    }
    // https://alpaca.markets/docs/api-references/market-data-api/stock-pricing-data/historical/#multi-bars
    async getBarsMulti(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrlData}${ALPACA_API_PATH_BARS_MULTI}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { query });
        }
    }
    // https://alpaca.markets/docs/api-references/market-data-api/stock-pricing-data/historical/#snapshot
    async getSnapshot(symbol) {
        try {
            const apiUrl = `${`${this.baseUrl}${ALPACA_API_PATH_SNAPSHOT}`.replace('{symbol}', symbol.toLowerCase())}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol });
        }
    }
    // https://alpaca.markets/docs/api-references/market-data-api/stock-pricing-data/historical/#multi-snapshots
    async getSnapshotMulti(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrl}${ALPACA_API_PATH_SNAPSHOT_MULTI}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, query);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/assets/
    async getAssets(query) {
        try {
            const apiUrl = getUrlWithQuery(`${this.baseUrl}${ALPACA_API_PATH_ASSETS}`, query);
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, query);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/assets/
    async getAsset(symbol) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_ASSETS}/${symbol}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/watchlist/
    async getWatchlists() {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_WATCHLISTS}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`);
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/watchlist/
    async getWatchlist(id) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_WATCHLISTS}/${id}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { id });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/watchlist/
    async createWatchlist(params) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_WATCHLISTS}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'POST',
                headers: this.headers,
                body: JSON.stringify(params),
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { params });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/watchlist/
    async updateWatchlist(id, params) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_WATCHLISTS}/${id}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'PUT',
                headers: this.headers,
                body: JSON.stringify(params),
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { params });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/watchlist/
    async addAssetToWatchlist(id, params) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_WATCHLISTS}/${id}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'POST',
                headers: this.headers,
                body: JSON.stringify(params),
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { params });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/watchlist/
    async deleteWatchlist({ id }) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_WATCHLISTS}/${id}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'DELETE',
                headers: this.headers,
            });
            return getResult(response, {
                type: 'emptyResponse',
            });
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { id });
        }
    }
    // https://alpaca.markets/docs/api-references/trading-api/watchlist/
    async removeAssetFromWatchlist(id, symbol) {
        try {
            const apiUrl = `${this.baseUrl}${ALPACA_API_PATH_WATCHLISTS}/${id}/${symbol}`;
            if (this.shouldLogUrl) {
                console.log('AlpacaClient', apiUrl);
            }
            const response = await fetch(apiUrl, {
                method: 'DELETE',
                headers: this.headers,
            });
            return getResult(response);
        }
        catch (error) {
            return getErrorJson(`Alpaca API fetch fail: ${error.message || 'unknown error'}`, { symbol });
        }
    }
}
