import moment from 'moment-timezone';
import { formatCurrencyNumber, formattedLog, getCustomComparisonName, getIndicators, getPercentChange, getTechnicalIndicatorsInput, } from '../../../lib/utils';
import { dojiCreator } from '../../../patterns/doji';
import configs from './config';
const doji = dojiCreator({ bodyPercentOfRange: 10 });
moment.tz.setDefault('America/New_York');
const { LOG_LEVEL = 'error' } = process.env;
const name = 'contradiption';
const shouldLog = LOG_LEVEL.includes(name);
const contradiption = async ({ bars, config, symbol, }) => {
    if (!bars?.length || bars.length < 4) {
        return;
    }
    const mostRecentBar = bars[bars.length - 1];
    if (moment(mostRecentBar.t).hours() !== 10 ||
        moment(mostRecentBar.t).minutes() !== 0) {
        return;
    }
    // opening bar should be a large drop
    const openingBar = bars[bars.length - 4];
    if (openingBar.o < openingBar.c) {
        return;
    }
    const openingPercentDrop = getPercentChange(openingBar.c, openingBar.o);
    if (openingPercentDrop < config.minPercentOpeningDrop) {
        if (shouldLog) {
            formattedLog({
                isBad: true,
                message: 'min percent drop not met',
                name,
                symbol,
            });
        }
        return;
    }
    const secondOpeningBar = bars[bars.length - 3];
    // second bar should be start a trend up and potentially be a doji
    if (secondOpeningBar.o > secondOpeningBar.c) {
        return;
    }
    if (config.shouldHaveDoji &&
        !doji(getTechnicalIndicatorsInput([secondOpeningBar]), symbol)) {
        if (shouldLog) {
            formattedLog({
                isBad: true,
                message: 'not a fitting doji',
                name,
                symbol,
            });
        }
        return;
    }
    const thirdOpeningBar = bars[bars.length - 2];
    // third bar should continue the upward trend
    if (thirdOpeningBar.o > thirdOpeningBar.c) {
        return;
    }
    // most recent bar should show a strong rise
    const mostRecentPercentRise = getPercentChange(mostRecentBar.o, mostRecentBar.c);
    if (mostRecentPercentRise < config.minPercentMostRecentRise) {
        if (shouldLog) {
            formattedLog({
                isBad: true,
                message: 'min percent rise not met',
                name,
                symbol,
            });
        }
        return;
    }
    const indicators = getIndicators({
        bars,
    });
    if (typeof config.minRvol === 'number' &&
        (!indicators.rvol || indicators.rvol < config.minRvol)) {
        return;
    }
    if (typeof config.maxRsi === 'number' &&
        config.maxRsi &&
        (!indicators.rsi || indicators.rsi > config.maxRsi)) {
        return;
    }
    const profitPercent = config.profitPercent;
    const lossPercent = config.lossPercent;
    const points = [mostRecentBar.t];
    const price = formatCurrencyNumber(mostRecentBar.c);
    const profitPrice = formatCurrencyNumber(price * (profitPercent / 100) + price);
    const stopPrice = formatCurrencyNumber(price - price * (lossPercent / 100));
    const entryTime = moment(mostRecentBar.t).format('YYYY-MM-DD h:mm:ss a');
    const trailPercent = !config.shouldUseTrailPercent
        ? undefined
        : (stopPrice / price) * 100;
    // custom comparisons
    const percentRiseName = 'percent-rise';
    const customComparisonPercentRise = getCustomComparisonName({
        increment: 0.1,
        name: percentRiseName,
        range: [0, 4],
        value: mostRecentPercentRise,
    });
    const rsiName = 'rsi';
    const customComparisonRsi = getCustomComparisonName({
        increment: 20,
        name: rsiName,
        range: [0, 100],
        value: indicators.rsi,
    });
    const rvolName = 'rvol';
    const customComparisonRvol = getCustomComparisonName({
        increment: 0.5,
        name: rvolName,
        range: [0, 10],
        value: indicators.rvol,
    });
    const customComparisons = [
        { group: percentRiseName, name: customComparisonPercentRise },
        { group: rsiName, name: customComparisonRsi },
        { group: rvolName, name: customComparisonRvol },
    ];
    return {
        ...indicators,
        customComparisons,
        name,
        points,
        price,
        profitPrice,
        stopPrice,
        t: mostRecentBar.t,
        time: entryTime,
        trailPercent,
    };
};
export { configs };
export default contradiption;
